--with Ada.Text_IO;

with Ticket_Automaton.Languages;

separate (Ticket_Automaton)
procedure Run is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.4
  -- Date      8 May 2022
  --====================================================================
  -- Starts and performs the automaton.
  --
  -- Revision was planned to be performed upon closing the ticket
  -- automaton, i.e. the window. This seems impossible, so it's not
  -- implemented and the corresponding texts are not translated.
  --
  -- This is what should have happened:
  -- When a purchase is being performed, it is abandoned with
  -- Message (Revision).
  -- Then a statistics is output about the accomplished purchases and
  -- the current number of coins and notes (Beagle_Boys).
  --====================================================================
  -- History
  -- Author Version   Date    Reason for Change
  --  C.G.    1.4  08.05.2022 Scrooge: Also count failures
  --  C.G.    1.3  15.04.2022 Comment about impossibility of Revision
  --  C.G.    1.2  13.04.2022 Help implemented
  --  C.G.    1.1  11.04.2022 Polyglotism
  --  C.G.    1.0  09.04.2022 Purchase done; missing Help, Language
  --                          selection, Revision (stop the automaton)
  --  C.G.    0.1  08.04.2022 Basic handling finished
  --  C.G.    0.0  02.04.2022 Template only (learning Ada_GUI)
  --====================================================================

  use Ada_GUI;
  package C renames Currency;

  subtype Coin_or_Note is Piece range C.Cent_10 .. C.Euro_10;
  subtype Denomination is C.Denomination (Coin_or_Note);
  Empty: constant Denomination  := (others => 0);

  Selected: Ticket_Choice := None;
  Inserted: Denomination  := Empty;

  Mouse: Next_Result_Info;  -- only mouse clicks

  package Scrooge is

    -- Change is returned in the smallest possible denomination; Change_Problem is False.
    -- Change_Problem is True if no matching denomination is possible; in this case,
    -- Change equals Paid.
    procedure Store (Selected      : in  Ticket_Kind;
                     To_be_Changed : in  Natural;
                     Paid          : in  Denomination;
                     Change        : out Denomination;
                     Change_Problem: out Boolean);

    type Tickets_Sold is array (Ticket_Choice) of Natural;

    -- Emptying the store
    procedure Beagle_Boys (Sales: out Tickets_Sold;
                           Store: out Denomination);

  end Scrooge;
  package body Scrooge is separate;

  package Dialogs is
    procedure Choose_Language;
    procedure Accomplished (Selected: Ticket_Kind ; Change: Denomination);
    procedure Abandonment  (Inserted: Denomination; Reason: String);
    procedure Instructions;
  end Dialogs;
  package body Dialogs is separate;

  use Languages;

begin

  loop
                                            -- Funny: Even
    Mouse := Next_Event (Timeout => 10.0);  -- background text
                                            -- creates events

    if Mouse.Timed_Out then
      if Selected /= None then
        Dialogs.Abandonment (Inserted, Reason => Message (Timeout).all);
        Selected := None;
        Ticket.Set_Text ("");
        Amount.Set_Text ("");
        Inserted := Empty;
      end if;
      goto Skip;
    end if;

    exit when Mouse.Event.Kind = Window_Closed;

    --Ada.Text_IO.Put_Line ("Event => " & '"' & Mouse.Event.ID.Text & '"');

    if Mouse.Event.ID = Help then  -- always selectable
      Dialogs.Instructions;
      goto Skip;
    end if;

    if Selected = None then  -- 1st step: select ticket

      if Mouse.Event.ID = Language then  -- only selectable before purchase start

        Dialogs.Choose_Language;

      else  -- serve only ticket button

        if Mouse.Event.ID = Short then
          Selected := Short_Trip;
        elsif Mouse.Event.ID = City then
          Selected := City_Region;
        elsif Mouse.Event.ID = Metropole then
          Selected := Metropole_Area;
        elsif Mouse.Event.ID = Day then
          Selected := Day_Ticket;
        else
          goto Skip;  -- wrong ID
        end if;

        Ticket.Set_Text (Mouse.Event.ID.Text);              -- display selection
        Amount.Set_Text (Image (Ticket_Price (Selected)));  -- and price ("&euro;" does not work)

      end if;

    else  -- insert coins until paid

      if Mouse.Event.ID = Cent_10 then
        Inserted (C.Cent_10) := @ + 1;
      elsif Mouse.Event.ID = Cent_20 then
        Inserted (C.Cent_20) := @ + 1;
      elsif Mouse.Event.ID = Cent_50 then
        Inserted (C.Cent_50) := @ + 1;
      elsif Mouse.Event.ID = Euro_1 then
        Inserted (C.Euro_1)  := @ + 1;
      elsif Mouse.Event.ID = Euro_2 then
        Inserted (C.Euro_2)  := @ + 1;
      elsif Mouse.Event.ID = Euro_5 then
        Inserted (C.Euro_5)  := @ + 1;
      elsif Mouse.Event.ID = Euro_10 then
        Inserted (C.Euro_10) := @ + 1;
      end if;

      if Mouse.Event.ID = Abortion then  -- cusomer changed mind, clear display
        Dialogs.Abandonment (Inserted, Reason => Message (Customer).all);
        Inserted := Empty;
        Selected := None;
        Ticket.Set_Text ("");
        Amount.Set_Text ("");
      else  -- continue purchase
        declare
          Too_Much: constant Integer := Ticket_Price (Selected) - C.Amount (Inserted);
          Change        :  Denomination;
          Change_Problem: Boolean;
        begin
          Amount.Set_Text (Image (Too_Much));
          if Too_Much <= 0 then  -- purchase done, clear display
            Scrooge.Store (Selected,
                           To_be_Changed => -Too_Much, Paid           => Inserted,
                           Change        => Change   , Change_Problem => Change_Problem);
            if Change_Problem then
              Dialogs.Abandonment (Inserted, Reason => Message (Change_Problems).all);
            else
              Dialogs.Accomplished (Selected, Change);
            end if;
            Selected := None;
            Inserted := Empty;
            Ticket.Set_Text ("");
          end if;
        end;
      end if;

    end if;

    <<Skip>>
  end loop;

  -- Implement Revision here upon closing window (if possible at all)

  Ada_GUI.End_GUI;

end Run;
