with Fahrkartenautomat.GUI;

package body Fahrkartenautomat is

  --====================================================================
  -- Autor     Christoph Grein
  -- Version   3.3
  -- Datum     1 April 2023
  --====================================================================
  -- Implementierung mit JEWL und Tasking.
  -- Identisch mit sequentieller Lsung bis auf Task Warte, die Hilfe
  -- und den Geldspeicher Dagobert.
  --
  -- 1. Fahrkarte whlen.
  --    Kann nicht wiederholt werden vor Beendigung des gesamten
  --    Vorgangs.
  --    Vor Wahl einer Karte ist die Geldeingabe oder Abbruch
  --    unmglich.
  --    Die gewhlte Fahrkarte und der zu zahlende Betrag werden
  --    angezeigt.
  --
  -- 2. Geld einwerfen.
  --    Kann nicht ausgefhrt werden vor Fahrkartenwahl und nach
  --    Beendigung des Vorgangs.
  --    Mit jeder Geldeingabe wird der zu zahlende Betrag angepasst.
  --    Der Vorgang wird beendigt, sobald gengend Geld eingeworfen
  --    ist oder wenn Abbruch gewhlt wird.
  --    a) Genug bezahlt
  --       Der zu zahlende Betrag wir gelscht, die gewhlte Fahrkarte
  --       wird weiterhin angezeigt.
  --       i)  Ein Infofenster erscheint mit Anzeige des Wechselgeldes.
  --           Sobald es besttigt wird, werden alle Anzeigen gelscht.
  --       ii) Wenn nicht genug Wechselgeld im Automaten ist, wird der
  --           Vorgang abgebrochen.
  --    b) Abbruch
  --       Alle Anzeigen werden gelscht.
  --       Ein Infofenster erscheint mit Anzeige des Rckgeldes.
  --    c) Zeitberschreitung
  --       Auerdem wird der Vorgang abgebrochen, wenn zu lange keine
  --       Eingabe erfolgt.
  --
  -- 3. Ausgabe der Fahrkarte mit Wechselgeld.
  --    Ende des Vorgangs.
  --    Bei Abbruch wird das eingeworfene Geld zurckerstattet und
  --    keine Fahrkarte ausgegeben.
  --====================================================================
  -- Geschichte
  -- Autor  Version  Datum    Grund der nderung
  --  C.G.    3.3  01.04.2023 Kartenwahl ausgrauen,
  --                          Geldtasten unsichtbar,
  --                          neues Kommando Zeitberschreitung
  --  C.G.    3.2  23.03.2023 Verkaufsstatistik neu
  --  C.G.    3.1  04.11.2020 Hilfe-Taste neu
  --  C.G.    3.0  03.11.2020 Funktional mit Tasking
  --  C.G.    2.0  03.11.2020 Funktional ohne Tasking
  --  C.G.    1.0  12.04.2014 Created for Ada tasking course
  --====================================================================

  package Dagobert is

    -- Das Wechselgeld wird in der kleinstmglichen Stckelung
    -- zurckgegeben; Wechselproblem ist False.
    -- Sollte eine passende Stckelung nicht mglich sein, ist das
    -- Wechselgeld das Gezahlte und Wechselproblem ist True.
    procedure Speichere (Gewhlt       : in  Fahrkarte ;
                         Zurckzuzahlen: in  Natural   ;
                         Gezahlt       : in  Stckelung;
                         Wechselgeld   : out Stckelung;
                         Wechselproblem: out Boolean   );

    function Panzerknacker return Stckelung;  -- Leerung des Speichers
    function Statistik     return Verkaufsstatistik;

  end Dagobert;

  package body Dagobert is separate;
  use Dagobert;

  use GUI;

  -- Zustand des Automaten

  Gewhlte_Karte: Fahrkarte;
  Zu_Zahlen     : Integer  := 0;  -- kann negativ werden bei berzahlung

  Eingegeben: Stckelung := (others => 0);

  procedure Whle_Fahrkarte (Kommando: in Kartenwahl) is
  begin
    case Kommando is
      when Kurzstrecke  => Gewhlte_Karte := Kurzstrecke;
      when Innenraum    => Gewhlte_Karte := Innenraum;
      when Gesamtgebiet => Gewhlte_Karte := Gesamtgebiet;
      when Tageskarte   => Gewhlte_Karte := Tageskarte;
    end case;
    Eingegeben := (others => 0);
    Zu_Zahlen  := Preis (Gewhlte_Karte);
    Anzeige_gewhlte_Karte  (Gewhlte_Karte'Image);
    Anzeige_Zu_Zahlen       (Zu_Zahlen);
    Aktiviere_Karten_Tasten (Aktiv    => False);  -- werden grau
    Zeige_Geld_Tasten       (Sichtbar => True );  -- erscheinen
  end Whle_Fahrkarte;

  procedure Abbruch (Ergebnis: out Ausgabe; Grund: in String);

  procedure Prfe_Summe (Kommando: in Bezahlen; Ergebnis: out Ausgabe) is
    Stck  : constant Geldstueck := Geldstueck'Val (GUI_Kommando'Pos (Kommando) - 2);
    Zurck : Stckelung;
    Problem: Boolean;
  begin
    Eingegeben (Stck) := Eingegeben (Stck) + 1;
    Zu_Zahlen          := Zu_Zahlen - Wert (Stck);
    Anzeige_Zu_Zahlen (Zu_Zahlen);
    if Zu_Zahlen <= 0 then
      Zeige_Geld_Tasten (Sichtbar => False);
      Speichere (Gewhlt        => Gewhlte_Karte,
                 Zurckzuzahlen => abs Zu_Zahlen,
                 Gezahlt        => Eingegeben,
                 Wechselgeld    => Zurck,
                 Wechselproblem => Problem);
      if Problem then
        Abbruch (Ergebnis, "wegen Wechselproblem");
        return;
      end if;
      Anzeige_Ausgabe         (abs Zu_Zahlen);  -- erst Dialog besttigen
      Anzeige_gewhlte_Karte  ("");             -- dann gewhlte Karte lschen
      Aktiviere_Karten_Tasten (Aktiv => True);  -- und neue Kartenwahl ermglichen
      Ergebnis     := (Gewhlte_Karte, Zurck);
      Eingegeben   := (others => 0);
    end if;
  end Prfe_Summe;

  procedure Abbruch (Ergebnis: out Ausgabe; Grund: in String) is
  begin
    Anzeige_gewhlte_Karte  ("");                          -- Reihenfolge wichtig!
    Anzeige_Zu_Zahlen       ( 0);                          -- Erst Anzeige anpassen, dann
    Zeige_Geld_Tasten       (Sichtbar => False);
    Anzeige_Abbruch         (Betrag (Eingegeben), Grund);  -- warten auf Besttigung im Dialog.
    Aktiviere_Karten_Tasten (Aktiv    => True);
    Ergebnis     := (Keine, Eingegeben);
    Eingegeben   := (others => 0);
  end Abbruch;

  procedure Zeige_Hilfe is
    -- Solange die Hilfe aktiv ist, muss die Eingabe unmglich sein,
    -- da die Anzeige nicht aktualisiert wird.
    Karten_Aktuell: constant Boolean := Karten_Tasten_Aktiv;
    Geld_Aktuell  : constant Boolean := Geld_Tasten_Sichtbar;
  begin
    Aktiviere_Karten_Tasten (Aktiv    => False);
    Zeige_Geld_Tasten       (Sichtbar => False);
    Bedienungsanleitung;
    Aktiviere_Karten_Tasten (Aktiv    => Karten_Aktuell);
    Zeige_Geld_Tasten       (Sichtbar => Geld_Aktuell);
  end Zeige_Hilfe;

  -----------------------------------------------------------------------------

  task Warte is
    entry Auf (Kommando: out GUI_Kommando);
  end Warte;

  task body Warte is
    Bereit: GUI_Kommando;
  begin
    Lebenslang:
    loop
      Bereit := Nchstes_Kommando;
      accept Auf (Kommando: out GUI_Kommando) do
        Kommando := Bereit;
      end Auf;
    end loop Lebenslang;
  end Warte;

  Wartezeit: constant array (Boolean) of Duration := (False => 10.0,            -- Zeitberschreitung wenn Automat nicht frei
                                                      True  => Duration'Last);  -- keine Zeitberschreitung mglich

  procedure Kaufe_Karte (Ergebnis: out Ausgabe; Stop: out Boolean) is
    Kommando: GUI_Kommando;
  begin
    Zeige_Geld_Tasten (Sichtbar => False);  -- zu Beginn nur Kartenwahl mglich
    loop
      select
        Warte.Auf (Kommando);
      or
        delay Wartezeit (Zu_Zahlen <= 0);
        Kommando := Zeitberschreitung;
      end select;
      case Kommando is
        when Hilfe              => Zeige_Hilfe;
        when Kartenwahl         => Whle_Fahrkarte (Kommando);
        when Bezahlen           => Prfe_Summe (Kommando, Ergebnis);
                                   exit when Zu_Zahlen <= 0;
        when Abbruch            |
             Zeitberschreitung => Abbruch (Ergebnis, (case Kommando = Abbruch is
                                                         when True  => "durch Kunden",
                                                         when False => "wegen Zeitberschreitung"));
                                   Zu_Zahlen := 0;  -- Automat frei wenn <= 0
                                   exit;
        when Ende               => Abbruch (Ergebnis, "wegen Revision");  -- kann immer passieren
                                   abort Warte;
                                   Anzeige_Speicherinhalt    (Panzerknacker);
                                   Anzeige_Verkaufsstatistik (Statistik);
                                   exit;
      end case;
    end loop;
    Stop := Kommando = Ende;
  end Kaufe_Karte;

end Fahrkartenautomat;
