------------------------------------------------------------------
--                                                              --
--                     This is Public Domain                    --
--                                                              --
-- This is not my invention. It has been derived from a         --
-- discussion started by Gautier de Montmollin on 20.02.2023 at --
--    https://groups.google.com/g/comp.lang.ada/c/3cTuOAdReio   --
-- I only changed the naming and added the documentation.       --
--                                                              --
------------------------------------------------------------------
-- For further documentation, see the project files.            --
------------------------------------------------------------------
--      This is the rather simplistic source code for the       --
--                       Hogwarts Compiler.                     --
------------------------------------------------------------------

with Ada.Calendar, Ada.Command_Line, Ada.Directories,
     Ada.Text_IO;
use  Ada.Text_IO;
with Ada.Unchecked_Deallocation;

procedure Hogwarts_Compiler is

  -- Definition of compiler switches (this is just an idea how to add
  -- them; they are just read but not used).
  -- A switch can have parameters (an arbirary sequence of graphic
  -- characters) after a colon; they are stored in the array Switch.
  -- There is a difference between a null switch (undefined) and one
  -- with an empty string as parameter (defined without parameter).

  type Switch_Name is (Snatch, Broom, Wand);

  type String_Access is access String;
  procedure Free is new Ada.Unchecked_Deallocation (String, String_Access);

  Switch: array (Switch_Name) of String_Access;

  function Index (X: String) return Positive is
    Haystack: constant String := X & ':';  -- so that the colon is always found
  begin
    for H in Haystack'Range loop
      if Haystack (H) = ':' then
        return H;
      end if;
    end loop;
  end Index;


  procedure Compile (Hogwarts_File: String) is

    -- hogwarts.gpr will assume the executable in directory Generated, this is
    -- also Current_Directory.
    -- Current_Directory: absolute\path\Generated
    -- Hogwarts_File    : absolute\path\Hogwarts_Code\hogwarts_castle.hog
    -- Hog_File_Name    : hogwarts_castle.hog
    -- Ada_Unit_Name    : hogwarts_castle
    -- Ada_File_Name    : hogwarts_castle.adb      will be in Generated
    Hog_File_Name: constant String := Ada.Directories.Simple_Name (Hogwarts_File);
    Ada_Unit_Name: constant String := Ada.Directories.Base_Name   (Hog_File_Name);
    Ada_File_Name: constant String := Ada_Unit_Name & ".adb";

    Hog_In, Ada_Out: File_Type;

    use type Ada.Calendar.Time;

  begin

    -- Put_Line (Ada.Directories.Current_Directory);
    -- Put_Line (Hogwarts_File);
    -- Put_Line (Hog_File_Name);
    -- Put_Line (Ada_Unit_Name);
    -- Put_Line (Ada_File_Name);

    if Ada.Directories.Exists (Ada_File_Name) and then
       Ada.Directories.Modification_Time (Ada_File_Name) >= Ada.Directories.Modification_Time (Hogwarts_File)
    then
      Put_Line ("Ada file """ & Ada_File_Name & """ is up to date");
      return;
    end if;

    Put ("Compiling """ & Hog_File_Name & """ into """ & Ada_File_Name & """...");

    Open   (Hog_In , In_File , Hogwarts_File);
    Create (Ada_Out, Out_File, Ada_File_Name);
    Put_Line (Ada_Out, "function " & ada_unit_name & " return String is");
    Put_Line (Ada_Out, "begin");
    Put_Line (Ada_Out, "  return """ & Get_Line (Hog_In) & """;");
    Put_Line (Ada_Out, "end " & ada_unit_name & ';');
    Close (Hog_In);
    Close (Ada_Out);

    Put_Line ("done");

  end Compile;

  use  Ada.Command_Line;

begin

  if Argument_Count = 0 then
    Put_Line (Current_Error, "Hogwarts file name missing");
    return;
  end if;

  -- Just to show how switches could be handled.
  -- Switches are transferred first, the compiler is the last argument.

  if Ada.Directories.Extension (Argument (Argument_Count)) /= "hog" then
    Put_Line (Current_Error, "Wrong file extension " & Argument (Argument_Count));
    return;
  end if;

  for A in 1 .. Argument_Count - 1 loop
    declare
      Arg: constant String   := Argument (A);
      Idx: constant Positive := Index (Arg);
    begin
      if Switch (Switch_Name'Value (Arg (1 .. Idx - 1))) /= null then
        Put_Line ("Switch " & Arg & " already defined, new value ignored");
      else
        Switch (Switch_Name'Value (Arg (1 .. Idx - 1))) := new String'(Arg (Idx + 1 .. Arg'Last));
        Put_Line (Arg (1 .. Idx - 1) & " => " & Arg (Idx + 1 .. Arg'Last));
      end if;
    exception
      when others =>
        Put_Line ("Unknown switch " & Arg & " ignored");
    end;
  end loop;

  Compile (Argument (Argument_Count));

  for S of Switch loop
    Free (S);  -- See RM 13.11.2(8)
  end loop;

end Hogwarts_Compiler;
