-- The so-called Beaujolais effect derives its name from a wager between
-- Jean Ichbiah and his team that there cannot be constructed an example
-- of a legal Ada program changing behaviour due to adding or removing a
-- use-clause.
-- Olivier Roubine found the first counterexample and did get his bottle
-- of Beaujolais.
-- This is an example of the effect (I do not know whether it is the
-- original one).
--
-- Jean D. Ichbiah:
-- "It is worth pointing out that many popular languages have Beaujolais
-- effect: e.g. the Borland Pascal 'uses' clause, which takes an
-- additive, layer after layer, interpretation of what you see in the
-- used packages (units), definitely exhibits a Beaujolais effect.
-- Last time I looked at C++, my impression was that several years of
-- Beaujolais vintage productions would be required.
-- For component-based software development, such effects are
-- undesirable since your application may stop working when you
-- recompile it with the new - supposedly improved - version of a
-- component."

package Beaujolais is
  procedure Drink (DOC: Boolean);  -- 1
end Beaujolais;
------------------------------------------------------------------------
with Text_IO;
use  Text_IO;

with Beaujolais;
use  Beaujolais;

procedure Beaujolais_Main is

  function "<" (L, R: Integer) return Integer is
  begin
    return R;
  end "<";

  procedure Drink (Vin_de_Table: Integer) is  -- 2
  begin
    Put_Line ("Land wine");
  end Drink;

begin

  -- Is this Drink 1 or 2 ?

  Drink (1 < 2);

  -- How is an expression like "1 < 2" interpreted?
  -- Ada83: RM 3.5.3(8) Integer literals are of the anonymous predefined
  --        type universal_integer.
  --        RM 4.6(15) says that implicit conversion is only done when
  --        there is no legal interpretation without.
  --        Thus with the use-clause present, there is a legal
  --        interpretation of "1 < 2" without type conversion, namely
  --        as an operation on universal_integer without implicit
  --        type conversion to Integer and so Drink 1 is called.
  --        When there is no use-clause, there is no interpretation
  --        without type conversion, so Drink 2 is called.
  --        This is the Beaujolais effect.
  -- Ada95: RM 3.5.4(14) Integer literals are all of the type
  --        universal_integer, the universal type for the class
  --        rooted at root_integer.
  --        RM 3.4.1(6) Universal types are analogous to class-wide
  --        types.
  --        RM 8.6(29) There is a preference for the primitive
  --        operators root numeric types root_integer and root_real.
  --        In particular, if two acceptable interpretations ...
  --        differ only in that one is for a primitive operator of
  --        the type root_integer or root_real, and the other is not,
  --        the interpretation using the primitive operator of the
  --        root numeric type is preferred.
  --        RM 3.4.1(12) shows why this preference is needed to resolve
  --        ambiguities.
  --        Thus the Boolean'(1 < 2) is interpreted as "<" on
  --        root_integer by preference, there is no implicit conversion
  --        of the literals.
  --        Without the use-clause, the boolean interpretation is not
  --        acceptable because it is not (directly) visible, the only
  --        acceptable remaining is Integer'(1 < 2) after implicit
  --        conversion of the literals. So Drink 2 is called.
  --        With the use-clause, both, the boolean and the integer
  --        interpretation are acceptable; there is no preference
  --        because the interpretations do not only differ in that...
  --        (the result types are different; so the expression is
  --        ambiguous.
  --        There is no Beaujolais effect.

end Beaujolais_Main;
------------------------------------------------------------------------
with Text_IO;
use  Text_IO;
package body Beaujolais is
  procedure Drink (Wine: Boolean) is
  begin
    Put_Line ("Beaujolais");
  end Drink;
end Beaujolais;