with System;

procedure Kaprekar_Number is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      13 November 2019
  --====================================================================
  -- Kaprekar Numbers
  --
  -- Definition
  -- https://en.wikipedia.org/wiki/Kaprekar_number
  -- (Note: This is the only place where 0 and Base**N are treated as
  -- Kaprekar numbers; in any other place, these numbers are excluded.)
  --
  -- Basically, if a natural number's square can be split into two
  -- parts such that the sum of the two parts is the number itself,
  -- this number is called a Kaprekar number.
  --
  -- Example base 10: 55**2 = 3025, 30 + 25 = 55
  -- Example base 16: CD**2 = A429, A4 + 29 = CD (205 in decimal)
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  13.11.2019 Created for public Ada course
  --====================================================================

  type Large_Integer is range System.Min_Int .. System.Max_Int;
  -- The maximum integer on this architecture.
  subtype Large_Natural     is Large_Integer range 0 .. Large_Integer'Last;
  subtype Large_Positive    is Large_Integer range 1 .. Large_Integer'Last;
  subtype Large_Number_Base is Large_Integer range 2 .. 16;

  procedure Put_Line (X: Large_Integer; Base: Large_Number_Base; Power: Positive) is separate;

  procedure Kaprekar (N: Large_Positive; Base: Large_Number_Base; Power: Positive; Left, Right: out Large_Natural) is
    -- Split the digits of the sqare of a natural number N into two parts Left and Right.
    -- Example: 55**2 = 3025 split into 30 and 25 for power 2.
    -- N is a Kaprekar number if Left + Right = N.
    Square : constant Large_Positive := N**2;
    Modulus: constant Large_Positive := Base**Power;
  begin
    Right := Square mod Modulus;
    Left  := (Square - Right) / Modulus;
  end Kaprekar;

  Base: constant Large_Number_Base := 13;

begin

  -- Let it run until N**2 overflows.
  for Power in Positive loop
    for N in Large_Positive range Base**(Power-1) .. Base**Power - 1 loop
      declare
        -- We need an inner loop.
        -- As an example see:
        -- N = 5292; N**2 = 28005264: Power = 4, Left = 2800, Right = 5264; Qower = 6, Left = 28, Right = 005264
        Qower: Positive := Power;
        Left, Right: Large_Natural;
      begin
        loop
          Kaprekar (N, Base, Qower, Left, Right);
          if Left + Right = N then
            Put_Line (N, Base, Qower);
            exit;
          end if;
          exit when Left = 0 or Left mod Base /= 0;
          Qower := Qower + 1;
        end loop;
      end;
    end loop;
  end loop;

end Kaprekar_Number;
