with Stack;

separate (UPN)
function Infix_to_Postfix (X: String) return String is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.1
  -- Date      30 January 2021
  --====================================================================
  -- Enclose the complete string by parentheses: A*B+D => (A*B+D).
  -- When comparing operator precedence, '(' is considered an operator
  -- with lowest precedence, and only there.
  -- Traverse all characters Z of Infix from left to right.
  -- # If Z is an opening parenthesis, push it on the stack (then the
  --   stack is no longer empty since this is the very first character).
  -- # If Z is a closing parenthesis, pop all operators down to the
  --   opening parenthesis and put them into Postfix.
  --   Discard both parentheses (so that in the end, the stack is empty
  --   again).
  -- # If Z is an operand, put it in Postfix.
  -- # If Z is an operator, pop all operators from the stack as long as
  --   their precedence is greater or equal to Z and put them in
  --   Postfix. (Thus an opening parenthesis stays in the stack.)
  --   Push Z on the stack.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.1  30.01.2021 Made it a subunit of UPN
  --  C.G.    1.0  28.01.2021 Created for Ada course
  --====================================================================

  subtype Extended is Character with Static_Predicate => Extended in Operator | '(' ;

  type Priority is range 0 .. 3;

  function Priority_of (Op: Extended) return Priority is
    (case Op is
       when '('       => 0,
       when '-' | '+' => 1,
       when '*' | '/' => 2,
       when '^'       => 3);

  Infix  : constant String := '(' & X & ')';
  Postfix: String (1 .. X'Length);
  Filled : Natural := 0;

begin

  for C of Infix loop
    case C is
      when Operand     => Filled := Filled + 1;
                          Postfix (Filled) := C;
      when Operator    => while Priority_of (C) <= Priority_of (Stack.Top) loop
                            Filled := Filled + 1;
                            Postfix (Filled) := Stack.Top;
                            Stack.Pop;
                          end loop;
                          Stack.Push (C);
      when '('         => Stack.Push (C);
      when ')'         => while Stack.Top in Operator loop
                            Filled := Filled + 1;
                            Postfix (Filled) := Stack.Top;
                            Stack.Pop;
                          end loop;
                          Stack.Pop;
      when others      => raise Program_Error;
    end case;
  end loop;

  return Postfix (1 .. Filled);

end Infix_to_Postfix;
