package Genetics is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      2 October 2019
  --====================================================================
  -- DNA replication may fail: Sometimes a wrong nucleotide is used.
  -- The Hamming distance counts the number of places where a wrong
  -- nucleotide is used for the replica.
  -- https://en.wikipedia.org/wiki/Hamming_distance
  --
  -- Example:
  --   "ACTAGCC"
  --   "ACGAGTC"
  --      ^  ^   Two differences, thus Hamming distance is 2.
  --
  -- There are several possibilities for denoting the four nucleotides
  -- Adenin, Cytosin, Guanin, Thymin.
  -- 1. You may use the predefined Character type 'A', 'C', 'G', 'T'
  --    so that a DNA strand becomes a String. This is the obvious
  --    choice but has the disadvantage that also other characters like
  --    'b' etc may be part of the string.
  -- 2. You may add a predicate:
  --    subtype DNA is String with
  --      Dynamic_Predicate =>
  --        (for all N of DNA => N in 'A' | 'C' | 'G' | 'T');
  --    This does not prevent other characters like 'b' from being part
  --    of the string, but they will raise an exception.
  -- 3. You may use an enumeration like
  --    type Nucleotide is (A, C, G, T);
  --    type DNA is array (Positive range <>) of Nucleotide;
  --    A DNA strand looks now like (A, T, T, A, G, C, ...), a rather
  --    clumsy notation.
  -- 4. You may use a character type, which is the choice used here.
  --    With this declaration, you may create DNA strands like Strings,
  --    however only these four characters are available.
  --    Note: In Ada, there is the predefined type Character and there
  --    are user-defined so-called character types.
  --    Thus the character 'A' of the character type Nucleotide has
  --    nothing to do with the predefined Character 'A'; the same goes
  --    for the three others (the internal representation is different).
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  02.10.2019 Created for Ada course
  --====================================================================

  type Nucleotide is ('A', 'C', 'G', 'T');

  type DNA is array (Positive range <>) of Nucleotide;

  function Hamming_Distance (Strand_1, Strand_2: DNA) return Natural
    with Pre => Strand_1'Length = Strand_2'Length and Strand_1'Length > 0;

end Genetics;
